
__all__ = [
    "JupyterAPIs",
    "DefaultOutputWidget"
]

class JupyterAPIs:
    """
    Provides access to the various Jupyter APIs
    """

    _apis = None
    @classmethod
    def load_api(cls):
        try:
            import IPython as base
        except ImportError:
            base = None

        try:
            import IPython.core.interactiveshell as shell
        except ImportError:
            shell = None
        try:
            import IPython.core.display as display
        except ImportError:
            display = None
        try:
            import ipywidgets as widgets
            widgets.Output
        except ImportError:
            widgets = None
        try:
            import ipyevents as events
        except ImportError:
            events = None

        cls._apis = (
            display,
            widgets,
            events,
            shell,
            base
        )

    @classmethod
    def get_shell_api(cls):
        if cls._apis is None:
            cls.load_api()
        return cls._apis[3]
    @classmethod
    def get_shell_instance(cls):
        return cls.get_shell_api().InteractiveShell.instance()

    @classmethod
    def get_base_api(cls):
        if cls._apis is None:
            cls.load_api()
        return cls._apis[4]
    @classmethod
    def in_jupyter_environment(cls):
        base = cls.get_base_api()
        if base is None:
            return False
        else:
            shell = base.get_ipython().__class__.__name__
            return 'zmq' in shell.lower()

    @classmethod
    def get_display_api(cls):
        if cls._apis is None:
            cls.load_api()
        return cls._apis[0]

    @property
    def display_api(self):
        return self.get_display_api()

    @classmethod
    def get_widgets_api(self):
        if self._apis is None:
            self.load_api()
        return self._apis[1]

    @property
    def widgets_api(self):
        return self.get_widgets_api()

    @classmethod
    def get_events_api(self):
        if self._apis is None:
            self.load_api()
        return self._apis[2]

    @property
    def events_api(self):
        return self.get_events_api()

    # @classmethod
    # def raw_html(cls, html):
    #     return cls.get_widgets_api().HTML(html)
    # @classmethod
    # def raw_markdown(cls, markdown):
    #     out = cls.get_widgets_api().Output()
    #     out.append_display_data(cls.get_display_api().Markdown(markdown))
    #     return out

class DefaultOutputWidget:
    _output_area_stack = []
    def __init__(self, obj=None):
        self.obj = JupyterAPIs.get_widgets_api().Output() if obj is None else obj
    @classmethod
    def get_default(cls):
        if len(cls._output_area_stack) == 0:
            return JupyterAPIs.get_widgets_api().Output()
        else:
            return cls._output_area_stack[-1]
    def __enter__(self):
        new = self.obj
        self._output_area_stack.append(new)
        return new
    def __exit__(self, exc_type, exc_val, exc_tb):
        self._output_area_stack.pop()
    def __call__(self, *args, **kwargs):
        return self.get_default()

class frozendict(dict):
    def __hash__(self):
        return hash(tuple(self.values())) + hash(tuple(self.keys()))
    def __setitem__(self, key, value):
        raise TypeError("`frozendict` is immutable")


import uuid

from .CommonCoordinateSystems import InternalCoordinateSystem, CartesianCoordinates3D
from .CoordinateSystemConverter import CoordinateSystemConverter
from ... import Numputils as nput
# import weakref

__all__ = [
    "GenericInternalCoordinateSystem",
    "GenericInternalCoordinates",
    "CartesianToGICSystemConverter",
    "GICSystemToCartesianConverter"
]

#TODO: these should all be metaclasses but :shrag:
class GenericInternalCoordinateSystem(InternalCoordinateSystem):
    """
    Represents ZMatrix coordinates generally
    """
    name = "GenericInternals"
    def __init__(self,
                 converter_options=None,
                 dimension=(None,),
                 coordinate_shape=(None,),
                 angle_ordering='ijk',
                 **opts):
        """
        :param converter_options: options to be passed through to a `CoordinateSystemConverter`
        :type converter_options: None | dict
        :param coordinate_shape: shape of a single coordinate in this coordiante system
        :type coordinate_shape: Iterable[None | int]
        :param dimension: the dimension of the coordinate system
        :type dimension: Iterable[None | int]
        :param opts: other options, if `converter_options` is None, these are used as the `converter_options`
        :type opts:
        """
        if converter_options is None:
            converter_options = opts
        converter_options['angle_ordering'] = converter_options.get('angle_ordering', angle_ordering)
        super().__init__(dimension=dimension, coordinate_shape=coordinate_shape, converter_options=converter_options)

    # def jacobian(self,
    #              coords,
    #              system,
    #              order=1,
    #              coordinates=None,
    #              converter_options=None,
    #              all_numerical=False,
    #              analytic_deriv_order=None,
    #              **finite_difference_options
    #              ):
    #     carts = self.convert_coords(coords, CartesianCoordinates3D,
    #                                 order=order,
    #                                 converter_options=converter_options
    #                                 )
    #     if system
    #     cart_jacs = carts.jacobian(system, order=order,
    #                                all_numerical=all_numerical,
    #                                analytic_deriv_order=analytic_deriv_order,
    #                                **finite_difference_options
    #                                )
    #     return nput.tensor_reexpand(
    #         super().jacobian(coords, CartesianCoordinates3D, order)
    #     )

GenericInternalCoordinates = GenericInternalCoordinateSystem()
GenericInternalCoordinates.__name__ = "ZMatrixCoordinates"
GenericInternalCoordinates.__doc__ = """Generic internals"""

class CartesianToGICSystemConverter(CoordinateSystemConverter):
    """
    A converter class for going from Cartesian coordinates to internals coordinates
    """

    @property
    def types(self):
        return (CartesianCoordinates3D, GenericInternalCoordinates)

    def convert_many(self, coords, *, specs, order=0, masses=None, remove_translation_rotation=True,
                     reference_coordinates=None,
                     return_derivs=None,
                     derivs=None,
                     gradient_function=None,
                     gradient_scaling=None,
                     **kw):
        """
        We'll implement this by having the ordering arg wrap around in coords?
        """

        if return_derivs is None or return_derivs is True:
            return_derivs = order
        if return_derivs == 0: return_derivs = 1
        if not nput.is_numeric(return_derivs):
            return_derivs = max(return_derivs)

        internals = nput.internal_coordinate_tensors(coords, specs, order=return_derivs, **kw)
        internals, derivs = internals[0], internals[1:]
        return internals, {
            'specs':specs,
            'derivs':derivs,
            'reference_coordinates':coords,
            'masses': masses,
            'remove_translation_rotation': remove_translation_rotation
        }

    def convert(self, coords, *, specs, order=0, **kw):
        return self.convert_many(coords, specs=specs, order=order, **kw)


class GICSystemToCartesianConverter(CoordinateSystemConverter):
    """
    A converter class for going from Cartesian coordinates to internals coordinates
    """

    @property
    def types(self):
        return (GenericInternalCoordinates, CartesianCoordinates3D)

    def convert_many(self, coords, *, reference_coordinates, specs, order=0,
                     masses=None,
                     remove_translation_rotation=True,
                     derivs=None,
                     return_derivs=None,
                     **kw):
        """
        We'll implement this by having the ordering arg wrap around in coords?
        """

        if return_derivs is None or return_derivs is True:
            return_derivs = order
        if return_derivs == 0: return_derivs = 1
        if not nput.is_numeric(return_derivs):
            return_derivs = max(return_derivs)
        (expansions, errors), _ = nput.inverse_coordinate_solve(specs, coords, reference_coordinates,
                                                                order=return_derivs,
                                                                return_expansions=True,
                                                                return_internals=True,
                                                                masses=masses,
                                                                remove_translation_rotation=remove_translation_rotation,
                                                                **kw
                                                                )
        carts, derivs = expansions[0], expansions[1:]
        return carts, {
            'specs':specs,
            'derivs': derivs,
            'masses': masses,
            'remove_translation_rotation': remove_translation_rotation
        }

    def convert(self, coords, *, reference_coordinates, specs, order=0, **kw):
        return self.convert_many(coords, reference_coordinates=reference_coordinates, specs=specs, order=order, **kw)

__converters__ = [CartesianToGICSystemConverter(), GICSystemToCartesianConverter()]




"""
Graphics primitives module
Provides stuff like Disk, Sphere, etc. and lets them figure out how to plot themselves
"""

__all__ = [
    "GraphicsPrimitive",
    "Sphere",
    "Cylinder",
    "Disk",
    "Line",
    "Text",
    "Arrow",
    "Inset",
    "Point",
    "Triangle",
    "Polygon"
]

import abc, numpy as np
from .VTKInterface import *

class GraphicsPrimitive(metaclass=abc.ABCMeta):
    @abc.abstractmethod
    def plot(self, axes, *args, graphics=None, **kwargs):
        """The one method that needs to be implemented, which takes the graphics and actually puts stuff on its axes

        :param axes:
        :type axes:
        :param args:
        :type args:
        :param kwargs:
        :type kwargs:
        :return:
        :rtype:
        """
        pass

    @abc.abstractmethod
    def get_bbox(self):
        ...

class Disk(GraphicsPrimitive):
    def __init__(self, position=(0, 0), radius=1, **opts):
        self.pos = position
        self.rad = radius
        self.opts = opts
        self.prim = None

    def get_bbox(self):
        return [(self.pos[0]-self.rad, self.pos[1]-self.rad), (self.pos[0]+self.rad, self.pos[1]+self.rad)]

    def plot(self, axes, *args, graphics=None, zdir=None, **kwargs):
        if hasattr(axes, 'axes'):
            axes = axes.axes

        return axes.draw_disk(self.pos, radius=self.rad, **dict(self.opts, **kwargs))
class Line(GraphicsPrimitive):
    def __init__(self, points, radius=.1, **opts):
        self.pos = points
        # self.pos2 = pos2
        # self.rest = rest
        self.rad = 72*radius # this can't be configured nearly as cleanly as the circle stuff...
        self.opts = opts
    @property
    def points(self):
        return self.pos
    def get_bbox(self):
        pos = np.array(self.points).T
        return [(np.min(pos[0]), np.min(pos[1])), (np.max(pos[0]), np.max(pos[1]))]
    def plot(self, axes, *args, graphics=None, **kwargs):
        if hasattr(axes, 'axes'):
            axes = axes.axes

        kw = dict(edgecolors=[[0.]*3+[.3]])
        kw = dict(kw, **self.opts)
        kw = dict(kw, s=[(10*self.rad)**2], **kwargs)
        return axes.draw_line(self.pos, **kw)

class Text(GraphicsPrimitive):
    def __init__(self, txt, pos, bbox=((1, 1), (1, 1)), **opts):
        self.txt = txt
        self.pos = pos
        self.bbox = bbox
        self.opts = opts
    def get_bbox(self):
        return [
            (self.pos[0]-self.bbox[0][0], self.pos[1]-self.bbox[1][0]),
            (self.pos[0]+self.bbox[0][1], self.pos[1]+self.bbox[1][1])
        ]
    def plot(self, axes, *args, graphics=None, **kwargs):
        if hasattr(axes, 'axes'):
            axes = axes.axes
        return axes.draw_text(self.pos, self.txt, **self.opts)

class Arrow(GraphicsPrimitive):
    def __init__(self, pos1, pos2, **opts):
        self.pos1 = pos1
        self.pos2 = pos2
        self.opts = opts
    def get_bbox(self):
        pos = np.array([self.pos1, self.pos2]).T
        return [(np.min(pos[0]), np.min(pos[1])), (np.max(pos[0]), np.max(pos[1]))]
    def plot(self, axes, *args, graphics=None, **kwargs):
        if hasattr(axes, 'axes'):
            axes = axes.axes
        return axes.draw_arrow([self.pos1, self.pos2], **self.opts)

class Sphere(GraphicsPrimitive):
    def __init__(self, position=(0, 0, 0), radius=1, sphere_points=48, **opts):
        self.pos = position
        self.rad = radius
        self.opts = dict(opts, sphere_points=sphere_points)

    def get_bbox(self):
        raise NotImplementedError("...")

    def plot(self, axes, *args, sphere_points=None, graphics=None, **kwargs):
        if hasattr(axes, 'axes'):
            axes = axes.axes
        return axes.draw_sphere(self.pos, self.rad, **self.opts)

class Cylinder(GraphicsPrimitive):
    def __init__(self, p1, p2, radius, circle_points = 32, **opts):
        self.pos1 = p1
        self.pos2 = p2
        self.rad = radius
        self.opts = opts
        self.circle_points = circle_points

    def get_bbox(self):
        raise NotImplementedError("...")

    def plot(self, axes, *args, circle_points=None, graphics=None, **kwargs):
        if isinstance(axes.figure, VTKWindow):
            self.prim = VTKCylinder(self.pos1, self.pos2, self.rad, **self.opts)
            s = self.prim
            return s.plot(axes.figure)
        else:
            if hasattr(axes, 'axes'):
                axes = axes.axes
            return axes.draw_cylinder(self.pos1, self.pos2, self.rad, **self.opts)

class Point(GraphicsPrimitive):
    def __init__(self, pts, **opts):
        self.pos = pts
        self.opts = dict(opts)

    def get_bbox(self):
        raise NotImplementedError("...")

    def plot(self, axes, *args, graphics=None, **kwargs):
        if hasattr(axes, 'axes'):
            axes = axes.axes
        return axes.draw_point(self.pos, **self.opts)

class Triangle(GraphicsPrimitive):
    def __init__(self, pts, **opts):
        self.pos = pts
        self.opts = dict(opts)

    def get_bbox(self):
        raise NotImplementedError("...")

    def plot(self, axes, *args, sphere_points=None, graphics=None, **kwargs):
        if hasattr(axes, 'axes'):
            axes = axes.axes
        return axes.draw_triangle(self.pos, **self.opts)

class Polygon(GraphicsPrimitive):
    def __init__(self, points, **opts):
        self.pos = points
        self.opts = dict(opts)

    def get_bbox(self):
        raise NotImplementedError("...")

    def plot(self, axes, *args, sphere_points=None, graphics=None, **kwargs):
        if hasattr(axes, 'axes'):
            axes = axes.axes
        return axes.draw_poly(self.pos, **self.opts)


class Inset(GraphicsPrimitive):
    def __init__(self, prims, position, offset=(.5, .5), dimensions=None, plot_range=None, **opts):
        self.prims = prims
        self.pos = position
        self.opts = opts
        self._plot_range = plot_range
        self._dimensions = dimensions
        self.offset = offset
        self._prim_cache = {}

    @property
    def plot_range(self):
        if self._plot_range is None:
            return self.get_plot_range()
        else:
            return self._plot_range
    @plot_range.setter
    def plot_range(self, pr):
        ((_, _), (_, _)) = pr
        self.plot_range = pr
    def get_plot_range(self):
        if len(self.prims) == 0:
            return [[0, 1], [0, 1]]
        [(plx, prx), (pby, pty)] = [[np.inf, -np.inf], [np.inf, -np.inf]]
        for g in self.prims:
            ((lx, by), (rx, ty)) = g.get_bbox()
            plx = min(lx, plx)
            prx = max(rx, prx)
            pby = min(by, pby)
            pty = max(ty,pty)
        return [(plx, prx), (pby, pty)]

    @property
    def dimensions(self):
        if self._dimensions is None:
            ((lx, rx), (by, ty)) = self.plot_range
            return (rx - lx, ty - by)
        else:
            dims = self._dimensions
            if dims[0] is None:
                ((lx, rx), (by, ty)) = self.plot_range
                w = (rx - lx)/(ty - by)*dims[1]
                dims = (w, dims[1])
            elif dims[1] is None:
                ((lx, rx), (by, ty)) = self.plot_range
                h = (ty - by)/(rx - lx)*dims[0]
                dims = (dims[0], h)
            return dims
    def get_bbox(self, graphics=None, preserve_aspect=None):
        w, h = self.dimensions
        if preserve_aspect is None and self._dimensions is not None:
            preserve_aspect = self._dimensions[0] is None or self._dimensions[1] is None
        if preserve_aspect and graphics is not None:
            ((lx, rx), (by, ty)) = graphics.plot_range
            gw = (rx - lx)
            gh = (ty - by)
            ar = graphics.aspect_ratio
            ((slx, srx), (sby, sty)) = self.plot_range
            sar = (sty - sby) / (srx - slx)
            if isinstance(ar, str) and ar == 'auto':
                w1, h1 = graphics.image_size
                ar = h1 / w1
            art = (gh/gw) / ar # the ratio of plot_range aspect to true aspect
            if self._dimensions[1] is None:
                h = w * (sar * art)
            else:
                w = h / (sar * art)

        ox, oy = self.offset
        x, y = self.pos
        bbox = [
            [x - ox * w, y - oy * h],
            [x + (1 - ox) * w, y + (1 - oy) * h],
        ]
        return bbox

    def get_axes(self, graphics, bbox=None, **opts):
        if bbox is None:
            bbox = self.get_bbox()
        if graphics.figure in self._prim_cache:
            self._prim_cache[graphics.figure].close()
        self._prim_cache[graphics.figure] = graphics.create_inset(bbox, **opts)
        return self._prim_cache[graphics.figure]

    def plot(self, axes, *args, graphics=None, **kwargs):
        if isinstance(axes.figure, VTKWindow):
            raise NotImplemented
        else:
            if graphics is None:
                graphics = axes
            bbox = self.get_bbox(graphics=graphics)
            g = self.get_axes(graphics, bbox, **self.opts)
            prims = [p.change_figure(g) if hasattr(p, 'change_figure') else p.plot(g) for p in self.prims]
            return prims

    # def __del__(self):
    #     if self._prim is not None:
    #         self._prim.remove()


import numpy as np
import itertools
from collections import namedtuple
from ..Data import UnitsData
from .. import Numputils as nput

__all__ = [
    "coordinate_label",
    "get_coordinate_label",
    "mode_label",
    "get_mode_labels",
    "coordinate_sorting_key",
    "sort_internal_coordinates",
    # "print_mode_labels"
]

atom_sort_order = ['C', 'O', 'N', 'H', 'D', 'F', 'Cl', 'Br', 'I']
coordinate_label = namedtuple('coordinate_label', ['ring', 'group', 'atoms', 'type'])
def get_coordinate_label(coord, atom_labels, atom_order=None):
    ring_lab = None
    group_lab = None
    # motif_lab = ""
    tag = None
    if atom_order is None:
        atom_order = atom_sort_order
    if not isinstance(atom_order, dict):
        atom_order = {a:i for i,a in enumerate(atom_order)}

    if isinstance(coord, dict):
        coord, tag = next(iter(coord.items()))
    if len(coord) == 2:
        if tag is None:
            tag = "stretch"

        a,b = coord
        atom_1 = atom_labels[a]
        atom_2 = atom_labels[b]
        r1 = atom_1.ring
        r2 = atom_2.ring
        if r1 is not None:
            if r2 is None or r1 == r2:
                ring_lab = r1
        elif r2 is not None:
            if r1 is None:
                ring_lab = r2

        g1 = atom_1.group
        g2 = atom_2.group
        if g1 is not None:
            if g2 is None or g1 == g2:
                group_lab = g1
        elif g2 is not None:
            if g1 is None:
                ring_lab = g2

        a1 = atom_1.atom
        a2 = atom_2.atom
        a,b = sorted([a1,a2], key=lambda c:atom_order.get(c, len(atom_order)))
        motif_lab = a+b

    elif len(coord) == 3:
        if tag is None:
            tag = "bend"

        a,b,c = coord
        atom_1 = atom_labels[a]
        atom_2 = atom_labels[b]
        atom_3 = atom_labels[c]

        r1 = atom_1.ring
        r2 = atom_2.ring
        r3 = atom_3.ring
        if (
                r1 is not None and r2 is not None and r3 is not None
                and r1 == r2 and r1 == r3 and r2 == r3
        ):
            ring_lab = r1

        g1 = atom_1.group
        g2 = atom_2.group
        g3 = atom_3.group
        if (
                g1 is not None and g2 is not None and g3 is not None
                and g1 == g2 and g1 == g3 and g2 == g3
        ):
            group_lab = g1

        a1 = atom_1.atom
        a2 = atom_2.atom
        a3 = atom_3.atom
        if atom_order.get(a1, len(atom_order)) > atom_order.get(a3, len(atom_order)):
            a1, a2, a3 = a3, a2, a1
        motif_lab = a1 + a2 + a3
    else:
        if tag is None:
            tag = "dihedral"

        a, b, c, d = coord
        atom_1 = atom_labels[a]
        atom_2 = atom_labels[b]
        atom_3 = atom_labels[c]
        atom_4 = atom_labels[d]

        rings = [atom_1.ring, atom_2.ring, atom_3.ring, atom_4.ring]
        if (
            all(r is not None for r in rings)
            and all(r1 == r2 for r1,r2 in itertools.combinations(rings, 2))
        ):
            ring_lab = rings[0]

        groups = [atom_1.group, atom_2.group, atom_3.group, atom_4.group]
        if (
                all(r is not None for r in groups)
                and all(r1 == r2 for r1, r2 in itertools.combinations(groups, 2))
        ):
            group_lab = groups[0]

        a1 = atom_1.atom
        a2 = atom_2.atom
        a3 = atom_3.atom
        a4 = atom_4.atom
        if atom_order.get(a1, len(atom_order)) > atom_order.get(a4, len(atom_order)):
            a1, a2, a3, a4 = a4, a3, a2, a1
        motif_lab = a1 + a2 + a3 + a4

    return coordinate_label(ring_lab, group_lab, motif_lab, tag)

coord_type_ordering = {'stretch': 0, 'bend': 1, 'dihedral': 2}
atom_coordinate_sort_order = ['H', 'D', 'O', 'N', 'F', 'Cl', 'Br', 'I', 'C']
def coordinate_sorting_key(label, type_ordering=None, atom_ordering=None):
    if type_ordering is None:
        type_ordering = coord_type_ordering

    if atom_ordering is None:
        atom_ordering = atom_coordinate_sort_order
    if not hasattr(atom_ordering, 'items'):
        atom_ordering = {a:i for i,a in enumerate(atom_ordering)}

    if isinstance(label, str): # assume it's just atoms
        return (len(label),) + tuple(sorted(atom_ordering.get(a, -1) for a in label))
    else:
        return (type_ordering.get(label.type, 3),) + tuple(sorted(atom_ordering.get(a, -1) for a in label.atoms))
def sort_internal_coordinates(coords, atoms=None, sort_key=None):
    if sort_key is None:
        sort_key = coordinate_sorting_key
    if atoms is not None:
        sort_key = lambda s: sort_key(
            s
                if not nput.is_int(s[0]) else
            "".join(atoms[i] for i in s)
        )
    if isinstance(coords, dict):
        return dict(
            sorted(coords.items(), key=lambda kv: sort_key(kv[1]))
        )
    else:
        return tuple(
            sorted(coords, key=sort_key)
        )

mode_label = namedtuple("mode_type", ["coefficients", "indices", "labels", "type"])
def get_mode_labels(
        internals,
        internal_modes_by_coords: np.ndarray,
        norm_cutoff=.8
):
    mode_coords = []
    if isinstance(internals, dict):
        indices = list(internals.keys())

        tags = list(internals.values())
    else:
        indices = list(range(len(internals)))
        tags = internals

    tags = [
        t.type
            if (
                hasattr(t, 'type')
                and t.type is not None
                and not isinstance(t.type, str)
            ) else
        t
        for t in tags
    ]

    for mode in internal_modes_by_coords.T:
        ord = np.argsort(-(mode**2))
        sort_mode = mode[ord]
        for i in range(len(mode)):
            sub_mode = sort_mode[:i+1]
            norm = np.linalg.norm(sub_mode)
            if norm > norm_cutoff:
                mode_coords.append(
                    [
                        sub_mode,
                        None if indices is None else [indices[j] for j in ord[:i+1]],
                        [tags[j] for j in ord[:i+1]]
                    ]
                )
                break
        else:
            mode_coords.append([
                sort_mode,
                None
                    if indices is None else
                [indices[j] for j in ord],
                [tags[j] for j in ord]
            ])

    types = []
    for coeffs, inds, tags in mode_coords:
        base_tags = tags[0]
        if isinstance(base_tags, str):
            base_tags = [base_tags]
        # check that no other modes have intruded
        keep = [True] * len(base_tags)
        for t in tags[1:]:
            if isinstance(t, str):
                t = [t]
            s = len(base_tags) - len(t)
            for k in range(s):
                keep[k] = False
            for i in range(len(t)):
                a = base_tags[s+i]
                b = t[i]
                if a is None or b is None or a != b:
                    keep[s+i] = False
        if any(
            kept and (bt is not None)
            for kept, bt in zip(keep, base_tags)
        ):
            subvec = [
                None if not kept else bt
                for kept, bt in zip(keep, base_tags)
            ]
            type_vec = type(base_tags)(*subvec)
        else:
            type_vec = None

        types.append(
            mode_label(coeffs, inds, tags, type_vec)
        )

    return types

# def print_mode_labels(modes, labels):
#     for i, (freq, lab) in enumerate(zip(reversed(modes.freqs), reversed(labels))):
#         print(
#             "Mode {} ({}): {:.0f} {}".format(i + 1,
#                                              len(modes.freqs) - (i+1),
#                                              freq * UnitsData.hartrees_to_wavenumbers,
#                                              "mixed"
#                                              if lab.type is None else
#                                              lab.type
#                                              )
#         )


"""
This lists the types of readers and things available to the GaussianLogReader
"""
import io

import numpy as np

from ....Parsers import *
from collections import namedtuple, OrderedDict

########################################################################################################################
#
#                                           OrcaLogComponents
#
# region OrcaLogComponents
Components = OrderedDict()  # we'll register on this bit by bit
# each registration should look like:

# Components["Name"] = {
#     "description" : string, # used for docmenting what we have
#     "tag_start"   : start_tag, # starting delimeter for a block
#     "tag_end"     : end_tag, # ending delimiter for a block None means apply the parser upon tag_start
#     "parser"      : parser, # function that'll parse the returned list of blocks (for "List") or block (for "Single")
#     "mode"        : mode # "List" or "Single"
# }

########################################################################################################################
#
#                                           CartesianCoordinates
#
input_start_tag = FileStreamerTag(
    """ Command line input:""",
)
input_end_tag = "----------------"
def parse_command_line(inp_str):
    return inp_str

Components["CommandLine"] = {
    "tag_start": input_start_tag,
    "tag_end"  : input_end_tag,
    "parser"   : parse_command_line,
    "mode"     : "Single"
}

calc_start_tag = FileStreamerTag(
    """Calculation info""",
    follow_ups=("----------------",)
)
calc_end_tag = "----------------"
def parse_command_line(inp_str):
    return inp_str

Components["CalculationInfo"] = {
    "tag_start": calc_start_tag,
    "tag_end"  : calc_end_tag,
    "parser"   : parse_command_line,
    "mode"     : "Single"
}

# region CartesianCoordinates

 # the region thing is just a PyCharm hack to collapse the boilerplate here... Could also have done 5000 files

cartesian_start_tag = FileStreamerTag(
    """Input structure:"""
)
cartesian_end_tag = "\n\n"

CRESTCoords = namedtuple("CRESTCoords", ["atoms", "coords"])
def cartesian_coordinates_parser(cart):
    coords = np.array(Number.findall(cart)).astype(float)
    atoms = Word.findall(cart)
    return CRESTCoords(
        atoms,
        coords.reshape(-1, 3)
    )

Components["InputStructure"] = {
    "tag_start": cartesian_start_tag,
    "tag_end"  : cartesian_end_tag,
    "parser"   : cartesian_coordinates_parser,
    "mode"     : "Single"
}


final_opt_info_start = "Final Geometry Optimization"
final_opt_info_end = "--------------------------"
def parse_opt_info(opt_block):
    return opt_block

Components["FinalOptInfo"] = {
    "tag_start": final_opt_info_start,
    "tag_end"  : final_opt_info_end,
    "parser"   : parse_opt_info,
    "mode"     : "Single"
}


final_ensemble_info = FileStreamerTag(
    "Final Ensemble Information",
    follow_ups=("--------------------------",)
)
final_ensemble_info_end = "-------"
EnsembleInfo = namedtuple('EnsembleInfo', ['relative_energies', 'total_energies', 'weights', 'report'])
def parse_ensemble_info(opt_block:str):
    header, engs = opt_block.split("Erel/kcal", 1)
    engs, footer = engs.rsplit('T /K', 1)
    report = header + 'T /K' + footer
    ensemble = np.loadtxt(io.StringIO(engs), skiprows=1, usecols=[1, 2, 3])
    return EnsembleInfo(
        *ensemble.T,
        report
    )

Components["FinalEnsembleInfo"] = {
    "tag_start": final_ensemble_info,
    "tag_end"  : final_opt_info_end,
    "parser"   : parse_ensemble_info,
    "mode"     : "Single"
}


########################################################################################################################
#
#                                           Defaults
#
# region Defaults
Defaults = (
)
# endregion

########################################################################################################################
#
#                                           Ordering
#
# region Ordering
# defines the ordering in the log file
glk = ( # this must be sorted by what appears when

)
list_type = { k:-1 for k in Components if Components[k]["mode"] == "List" }
Ordering = { k:i for i, k in enumerate([k for k in glk if k not in list_type]) }
Ordering.update(list_type)
del glk
del list_type
# endregion

__components__ = Components
__ordering__ = Ordering
__defaults__ = Defaults



"""
Provides utilities for managing object persistence.
Two classes of persistence are provided.
 1. Config persistence: stores objects by turning them into a
    set of config variables & provides reloading
 2. File-backed objects: stores objects by making serializing core
    pieces of the data
"""
import json
import os, shutil, tempfile as tf, weakref

from .Checkpointing import Checkpointer, NumPyCheckpointer
from .Configurations import Config

__all__ = [
    "PersistenceLocation",
    "PersistenceManager",
    "ResourceManager"
]

class PersistenceLocation:
    """
    An object that tracks a location to persist data
    and whether or not that data should be cleaned up on
    exit
    """
    _cache = weakref.WeakValueDictionary()
    def __init__(self, loc, name=None, delete=None):
        if name is None:
            name = os.path.basename(loc)
        self.name = name

        if delete is None:
            delete = not os.path.isdir(loc)

        absloc = os.path.abspath(loc)
        if not os.path.isdir(absloc):
            if absloc != loc:
                loc = os.path.join(tf.TemporaryDirectory().name, loc)
        else:
            loc = absloc

        # all it takes is a single location
        # saying "don't delete" for us to not
        # delete...note that if the location
        # dies and then is reborn as a deletable
        # then it will be deleted
        for k,v in self._cache.items():
            if v.loc == loc:
                if not delete:
                    v.delete = False
                elif not v.delete:
                    delete = False

        self.loc = loc
        self.delete = delete

        self._cache[loc] = self

    def __repr__(self):
        return "{}({}, {}, delete={})".format(
            type(self).__name__,
            self.name,
            self.loc,
            self.delete
        )

    def __del__(self):
        if self.delete:
            try:
                shutil.rmtree(self.loc)
            except OSError:
                pass

class PersistenceManager:
    """
    Defines a manager that can load configuration data from a directory
    or, maybe in the future, a SQL database or similar.
    Requires class that supports `from_config` to load and `to_config` to save.
    """
    def __init__(self, cls, persistence_loc=None):
        """
        :param cls:
        :type cls: type
        :param persistence_loc: location from which to load/save objects
        :type persistence_loc: str | None
        """
        self.cls = cls
        if persistence_loc is None:
            if not hasattr(cls, 'persistence_loc'):
                raise AttributeError("{me}: to support persistence either a '{ploc}'' must be passed or {cls} needs the attribute '{ploc}'".format(
                    me=type(self).__name__,
                    cls=cls.__name__,
                    ploc=persistence_loc
                ))
            persistence_loc = cls.persistence_loc
        self.loc = persistence_loc

    def obj_loc(self, key):
        return os.path.join(self.loc, key)

    def load_config(self, key, make_new=False, init=None):
        """
        Loads the config for the persistent structure named `key`
        :param key:
        :type key:
        :return:
        :rtype:
        """
        if self.contains(key):
            return Config(self.obj_loc(key), extra_params=init)
        elif make_new:
            return self.new_config(key, init=init)
        else:
            raise KeyError("{}: no persistent object {}".format(
                type(self).__name__,
                key
            ))

    def new_config(self, key, init=None):
        """
        Creates a new space and config for the persistent structure named `key`

        :param key: name for job
        :type key: str
        :param init: initial parameters
        :type init: str | dict | None
        :return:
        :rtype:
        """
        loc = self.obj_loc(key)
        if init is None:
            init = {}
        elif isinstance(init, str):
            if os.path.isdir(init):
                init_dir = init
                init = Config(init).opt_dict
                if 'initialization_directory' not in init:
                    init['initialization_directory'] = init_dir
            else:
                init = Config(init).opt_dict

        if not os.path.isdir(loc):
            if 'initialization_directory' in init:
                shutil.copytree(
                    init['initialization_directory'],
                    loc
                )
            else:
                os.makedirs(loc, exist_ok=True)

        # now we prune, because we don't want to preserve this forever...
        if 'initialization_directory' in init:
            del init['initialization_directory']

        if Config.find_config(loc) is None:
            return Config.new(loc, init=init)
        else:
            conf = Config(loc)
            if len(init) > 0:
                conf.update(**init)
            return conf

    def contains(self, key):
        """
        Checks if `key` is a supported persistent structure

        :param key:
        :type key:
        :return:
        :rtype:
        """
        return os.path.isdir(self.obj_loc(key))

    def load(self, key, make_new=False, strict=True, init=None):
        """
        Loads the persistent structure named `key`

        :param key:
        :type key:
        :return:
        :rtype:
        """
        cfg = self.load_config(key, make_new=make_new, init=init)
        try:
            loader = self.cls.from_config
        except AttributeError:
            loader = None
        if loader is None:
            raise AttributeError("{}.{}: to support persistence {} has to have a classmethod '{}'".format(
                type(self).__name__,
                'load',
                self.cls,
                'from_config'
            ))
        return cfg.apply(loader, strict=strict)
    def save(self, obj):
        """
        Saves requisite config data for a structure

        :param obj:
        :type obj:
        :return:
        :rtype:
        """
        if not isinstance(obj, self.cls):
            raise TypeError("{}.{}: object {} isn't of persistence type {}".format(
                type(self).__name__,
                'save',
                obj,
                self.cls.__name__
            ))

        try:
            loader = obj.to_config
        except AttributeError:
            loader = None
        if loader is None:
            raise AttributeError("{}.{}: to support persistence {} has to have a classmethod '{}'".format(
                type(self).__name__,
                'save',
                self.cls,
                'to_config'
            ))
        data = loader()

        key = data['name']
        cfg = self.load_config(key, make_new=True)
        cfg.update(**data)

class ResourceManager:
    """
    A very simple framework for writing resources to a given directory
    Designed to be extended and to support metadata
    """

    default_resource_name = 'resource'
    def __init__(self, name=None, location=None, write_metadata=False, temporary=None):
        self.name = self.default_resource_name if name is None else name
        self.location = self.get_base_location(temporary=temporary) if location is None else location
        self.write_metadata = write_metadata

    base_location = None
    location_env_var = None
    use_temporary = True
    @classmethod
    def resolve_shared_directory(cls):
        #TODO: make this less Unix specific
        return os.path.expanduser('~/.local')
    @classmethod
    def get_default_base_location(cls, temporary=None):
        if temporary:
            return tf.TemporaryDirectory().name
        else:
            return cls.resolve_shared_directory()
    @classmethod
    def get_base_location(cls, temporary=True):
        if cls.base_location is None:
            if cls.location_env_var is not None:
                cls.base_location = os.environ.get(cls.location_env_var)
            if cls.base_location is None:
                cls.base_location = cls.get_default_base_location(temporary=temporary)
        return cls.base_location

    def get_resource_path(self, *path):
        return os.path.join(self.location, self.name, *path)

    blacklist_files = ['.DS_Store']
    def list_resources(self):
        base_dir = self.get_resource_path()
        os.makedirs(base_dir, exist_ok=True)
        return {
            p:os.path.join(base_dir, p)
            for p in os.listdir(base_dir)
            if p not in self.blacklist_files
        }

    binary_resource = True
    json_resource = False
    def save_resource(self, loc, val):
        with open(loc, 'w+' if not self.binary_resource else 'wb') as res_file:
            if self.json_resource:
                json.dump(res_file, val)
            else:
                res_file.write(val)
    def load_resource(self, loc):
        with open(loc, 'r' if not self.binary_resource else 'rb') as res_file:
            if self.json_resource:
                return json.load(res_file)
            else:
                return res_file.read()

    def get_metadata_filename(self, name):
        return name + '.meta.json'
    def get_resource_metadata(self, loc):
        return {}
    def get_resource_filename(self, name):
        return name

    resource_function = None
    def get_resource(self, name,
                     resource_function=None,
                     load_resource=True):
        resource_file = self.get_resource_filename(name)
        loc = self.get_resource_path(resource_file)
        if not os.path.exists(loc):
            base_dir = self.get_resource_path()
            os.makedirs(base_dir, exist_ok=True)
            if resource_function is None:
                resource_function = self.resource_function
            if resource_function is not None:
                default_val = resource_function(name)
                self.save_resource(loc, default_val)
            else:
                return None
        if load_resource:
            return self.load_resource(loc)
        else:
            return loc